import matplotlib.pyplot as plt

def format(xlab, ylab, legend=False, show=False, fileName=False):
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.xlabel(xlab, fontsize=18)
    plt.ylabel(ylab, fontsize=18)
    plt.grid()
    plt.tight_layout()
    if legend[0]:
        plt.legend()
    if fileName:
        plt.savefig(fileName + '.png')
    if show:
        plt.show()

def plot(x, y, xlab='$E$ / V', ylab='$i$ / A', mark='-', legend=False, fig=1, show=False, fileName=False):
    '''
        Function to plot curves without all the hassle. By default, y is a list 
        and mark is a string. To use different markers for each curve, include them as 
        a list where each item is a different marker / line style. y can also be a
        numpy array.

        Example:
        > import numpy as np
        > x = np.array([1,2,3,4])
        > y = [x, x**2, x***3]
        > mark = ['-', '--', 'or-']
        > legend = ['x', 'x^2', 'x^3']
        > sp.plotting.plot(x, y, xlab='x', ylab='y', mark=mark, legend=legend, show=1)
    '''
    # By default, y should be a list, if it is just a numpy array then
    # convert it into list to be used in the for loop afterwards:
    if (type(y) != list):
        y = [y]
    ny = len(y)
    # Even with multiple plots, the user may want to have the same marker or
    # line stile for each curve, so if it is not a list, convert it into one
    if (type(mark) != list):
        mark = [mark]*ny
    # Legend can also be just one or a list:
    if (type(legend) != list):
        legend = [legend]*ny
    plt.figure(fig)
    for n in range(ny):
        plt.plot(x, y[n], mark[n], label=legend[n])
    
    format(xlab, ylab, legend, show, fileName)


if __name__ == '__main__':
    import numpy as np
    x = np.array([1,2,3,4])
    y = [x, x**2, x**3]
    mark = ['-vg', '--b', '-or']
    legend = ['x', 'x^2', 'x^3']
    plot(x, y, xlab='x', ylab='y', mark=mark, legend=legend, show=1)



"""
Soft Potato pip version

"""

from .calculate import *
from .simulate import *
from .plotting import *
from .technique import *

__version__ = "1.0.14"
__author__ = 'Oliver Rodriguez'

__all__ = ['technique', 'calculate', 'plotting', 'simulate']



import numpy as np
from scipy.sparse import diags

'''
    Pending:
    - Add C mechanism
    - Calculate concentration of P on the EC and C mechanisms

    Errors:
    - Denormalization of concentrations is wrong
'''

F = 96485 # C/mol, Faraday constant    
R = 8.315 # J/mol K, Gas constant
T = 298 # K, Temperature
FRT = F/(R*T)

class E:
    '''
        Defines an E species
    '''
    def __init__(self, n=1, DO=1e-5, DR=1e-5, cOb=0, cRb=1e-6, E0=0, ks=1e8, 
                 alpha=0.5):
        self.n = n
        self.DO = DO
        self.DR = DR
        self.DOR = DO/DR
        self.cOb = cOb
        self.cRb = cRb
        self.E0 = E0
        self.ks = ks
        self.alpha = alpha


class C:
    '''
        Defines a C species  
    '''
    def __init__(self, DP=1e-5, cPb=0, kc=1e-2):
        self.DP = DP
        self.cPb = cPb
        self.kc = kc
        #self.Kc = 0


class TGrid:
    '''
        Defines the grid in time
    '''
    def __init__(self, twf, Ewf):
        self.t = twf
        self.E = Ewf
        self.nT = np.size(self.t) # number of time elements
        self.dT = 1/self.nT # adimensional step time
        self.T = twf/twf[-1] # adimensional time

        
class XGrid:
    '''
        Defines the grid in space
    '''
    def __init__(self, species, tgrid, Ageo=1):
        self.lamb = 0.45 # For the algorithm to be stable, lamb = dT/dX^2 < 0.5
        self.Xmax = 6*np.sqrt(tgrid.nT*self.lamb) # Infinite distance
        self.dX = np.sqrt(tgrid.dT/self.lamb) # distance increment
        self.nX = int(self.Xmax/self.dX) # number of distance elements
        self.X = np.linspace(0, self.Xmax, self.nX) # Discretisation of distance
        self.Ageo = Ageo

        for x in species:
            if isinstance(x, E):
                ## Discretisation of variables and initialisation
                if x.cRb == 0: # In case only O present in solution
                    x.CR = np.zeros([tgrid.nT, self.nX])
                    x.CO = np.ones([tgrid.nT, self.nX])
                else:
                    x.CR = np.ones([tgrid.nT, self.nX])
                    x.CO = np.ones([tgrid.nT, self.nX])*x.cOb/x.cRb

                x.eps = (tgrid.E-x.E0)*x.n*FRT # adimensional potential waveform
                x.delta = np.sqrt(x.DR*tgrid.t[-1]) # cm, diffusion layer thickness
                x.Ke = x.ks*x.delta/x.DR # Normalised standard rate constant
                # Construct matrix:
                Cb = np.ones(self.nX-1) # Cbefore
                Cp = -2*np.ones(self.nX) # Cpresent
                Ca = np.ones(self.nX-1) # Cafter
                x.A = diags([Cb,Cp,Ca], [-1,0,1]).toarray()/(self.dX**2) 
                x.A[0,:] = np.zeros(self.nX)
                x.A[0,0] = 1 # Initial condition
            elif isinstance(x, C):
                x.CP = np.zeros([tgrid.nT, self.nX])


class Simulate:
    '''
    '''
    def __init__(self, species, mech, tgrid, xgrid):
        self.species = species
        self.mech = mech
        self.tgrid = tgrid
        self.xgrid = xgrid

        self.join(species)

    def sim(self):
        nE = self.nE[0]
        nC = self.nC[0]
        sE = self.species[nE]
        sC = self.species[nC]
        for k in range(1, self.tgrid.nT):
            # Boundary condition, Butler-Volmer:
            CR1kb = sE.CR[k-1,1]
            CO1kb = sE.CO[k-1,1]
            sE.CR[k,0] = (CR1kb + self.xgrid.dX*sE.Ke*np.exp(-sE.alpha*sE.eps[k]
                         )*(CO1kb + CR1kb/sE.DOR))/(1 + self.xgrid.dX*sE.Ke*(
                         np.exp((1-sE.alpha)*sE.eps[k])+np.exp(
                         -sE.alpha*sE.eps[k])/sE.DOR))
            sE.CO[k,0] = CO1kb + (CR1kb - sE.CR[k,0])/sE.DOR
            # Runge-Kutta 4:
            sE.CR[k,1:-1] = self.RK4(sE.CR[k-1,:], 'E', sE)[1:-1]
            if self.mech == 'E':
                sE.CO[k,1:-1] = self.RK4(sE.CO[k-1,:], 'E', sE)[1:-1]
            elif self.mech == 'EC':
                sE.CO[k,1:-1] = self.RK4(sE.CO[k-1,:], 'EC', sE, sC)[1:-1]
                sC.CP[k,0] = sC.CP[k-1,1]
                sC.CP[k,1:-1] = self.RK4(sC.CP[k-1,:], 'ECP', sE, sC)[1:-1]
                
        for s in self.species:
            if isinstance(s, E):
                # Denormalising:
                if s.cRb:
                    I = -s.CR[:,2] + 4*s.CR[:,1] - 3*s.CR[:,0]
                    D = s.DR
                    c = s.cRb
                else: # In case only O present in solution
                    I = s.CO[:,2] - 4*s.CO[:,1] + 3*s.CO[:,0]
                    D = s.DO
                    c = s.cOb
                self.i = s.n*F*self.xgrid.Ageo*D*c*I/(2*self.xgrid.dX*s.delta)

                #self.cR = s.CR*s.cRb
                #self.cO = s.CO*s.cOb
                self.x = self.xgrid.X*s.delta
            elif isinstance(s, C):
                self.cP = s.CP#*s.cPb

        s.CR = 0 # Empty matrix to free memory
        s.CO = 0
        s.CP = 0

    def join(self, species):
        ns = len(species)
        self.nE = []
        self.nC = []
        for s in range(ns):
            if isinstance(species[s], E):
                self.nE.append(s)
            elif isinstance(species[s], C):
                self.nC.append(s)
        # User may only want to simulate an E mechanism:
        if not self.nC:
            self.nC.append(0)
        if not self.nE:
            self.nE.append(0)

        if self.mech == 'EC' or self.mech == 'C':
            species[self.nC[0]].Kc = species[self.nC[0]].kc* \
                                     species[self.nE[0]].delta/ \
                                     species[self.nE[0]].DR

    def fun(self, y, mech, species, params):
        rate = 0
        if mech == 'EC':
            rate = - self.tgrid.dT*params.Kc*y
        elif mech == 'ECP': # To obtain concentration profile of P
            rate = self.tgrid.dT*params.Kc*y
        return np.dot(species.A, y) + rate

    def RK4(self, y, mech, species, params=0):
        dT = self.tgrid.dT
        k1 = self.fun(y, mech, species, params)
        k2 = self.fun(y+dT*k1/2, mech, species, params)
        k3 = self.fun(y+dT*k2/2, mech, species, params)
        k4 = self.fun(y+dT*k3, mech, species, params)
        return y + (dT/6)*(k1 + 2*k2 + 2*k3 + k4)



if __name__ == '__main__':
    from calculate import *
    from technique import *
    from plotting import *

    # CV
    wf = Sweep(Eini=0.5, Efin=-0.5, sr=0.01)
    twf = wf.t
    Ewf = wf.E
    e = E(cRb=0, cOb=1e-6)
    tgrid = TGrid(twf, Ewf)
    xgrid = XGrid([e], tgrid)
    simE = Simulate([e], 'E', tgrid, xgrid)
    simE.sim()
    e = E(cRb=0, cOb=1e-6)
    c = C(kc=1e-1)
    xgrid = XGrid([e,c], tgrid)
    simEC = Simulate([e,c], 'EC', tgrid, xgrid)
    simEC.sim()
    plot(Ewf, [simE.i, simEC.i], xlab='$E$ / V', ylab='$i$ / A', 
         legend=['E', 'EC'], fig=1, show=0)

    # CA
    wf = Step(Es=-0.4, dt=0.01)
    twf = wf.t
    Ewf = wf.E
    tgrid = TGrid(twf, Ewf)
    e = E(cRb=0, cOb=1e-6)
    xgrid = XGrid([e], tgrid)
    simE = Simulate([e], 'E', tgrid, xgrid)
    simE.sim()
    e = E(cRb=0, cOb=1e-6)
    c = C(kc=1e-1)
    xgrid = XGrid([e,c], tgrid)
    simEC = Simulate([e,c], 'EC', tgrid, xgrid)
    simEC.sim()

    # Cottrell
    macro = Macro()
    i_cott = macro.Cottrell(twf)

    legend = ['E', 'EC', 'Cottrell']
    mark = ['-o', 'o', '--']
    plot(1/np.sqrt(twf[1:]), [simE.i[1:], simEC.i[1:], i_cott[1:]], 
         xlab='$t^{-1/2}$ / s$^{-1/2}$', ylab='$i$ / A', 
         legend=legend, mark=mark, fig=2, show=0)
    plot(twf[1:], [simE.i[1:]/i_cott[1:], simEC.i[1:]/i_cott[1:]],
         xlab='$t$ / s', ylab='$i$ / $i_{cott}$',
         legend=legend, mark=mark, fig=3, show=1)



import numpy as np

class Sweep:
    """

    Returns t and E for a sweep potential waveform.
    All the parameters are given a default value.

    ----------
    Parameters:
    Eini:   V, initial potential [-0.5 V]
    Efin:   V, final potential [0.5 V]
    sr:     V/s, scan rate [0.1 V/s]
    dE:     V, potential increment [0.01 V]
    ns:     number of sweeps [2]

    ----------
    Returns:
    self.t: s, time numpy array
    self.E: V, potential numpy array

    ----------
    Examples:
    > import softpotato as sp
    > cv = sp.technique.Sweep(sr=0.001) # Changing only the scan rate
    print(cv.E)
    print(cv.t)

    """

    def __init__(self, Eini=-0.5, Efin=0.5, sr=0.1, dE=0.01, ns=2):
        self.Eini = Eini
        self.Efin = Efin
        self.sr = sr
        self.dE = dE
        self.ns = ns

        Ewin = abs(self.Efin-self.Eini)
        tsw = Ewin/self.sr # total time for one sweep
        nt = int(Ewin/self.dE)

        self.E = np.array([])
        self.t = np.linspace(0, tsw*self.ns, nt*self.ns)

        for n in range(1, self.ns+1):
            if (n%2 == 1):
                self.E = np.append(self.E, np.linspace(self.Eini, self.Efin, nt))
            else:
                self.E = np.append(self.E, np.linspace(self.Efin, self.Eini, nt))

class Sweep2:
    """ 

    Returns t and E for a sweep potential waveform.
    All the parameters are given a default value.

    ----------
    Parameters:
    Eini:   V, initial potential [-0.5 V]
    Efin:   V, final potential [0.5 V]
    sr:     V/s, scan rate [0.1 V/s]
    dE:     V, potential increment [0.01 V]
    ns:     number of sweeps [2]

    ----------
    Returns:
    self.t: s, time numpy array
    self.E: V, potential numpy array

    ----------
    Examples:
    > import softpotato as sp
    > cv = sp.technique.Sweep(sr=0.001) # Changing only the scan rate
    print(cv.E)
    print(cv.t)

    """

    def __init__(self, Eini=0, Eupp=0.5, Elow=-0.5, sr=0.1, dE=0.01, ns=2): # Modified class to include upper and lower vertex potentials
        self.Eini = Eini
        self.Eupp = Eupp # Changed the Efin to upper vertex potential
        self.Elow = Elow # Added a lower vertex potential
        self.sr = sr
        self.dE = dE
        self.ns = ns

        Ewin = abs(self.Eupp-self.Elow) # Adapted Ewin to accomodate upper and lower vertex potentials
        tsw = Ewin/self.sr # total time for one sweep
        nt = int(Ewin/self.dE)
        ntupper = int(((Eupp - Eini)/Ewin)*nt) # Added ntupper and ntlower parameters to allow for asymmetric waveforms
        ntlower = int(((Eini - Elow)/Ewin)*nt)

        self.E = np.array([])
        self.t = np.linspace(0, tsw*self.ns, nt*self.ns)

        if Eini == Eupp or Eini == Elow: # Accomodates for situations in which initial potential is same as uppoer or lower vertex
            segments = 2
        else:
            segments = 3
            
        for n in range(1, self.ns+1):
            if (segments == 3) and (abs(dE) == dE):
                self.E = np.append(self.E, np.linspace(self.Eini, self.Eupp, ntupper))
                self.E = np.append(self.E, np.linspace(self.Eupp, self.Elow, nt))
                self.E = np.append(self.E, np.linspace(self.Elow, self.Eini, ntlower))
            elif (segments == 3) and (abs(dE) != dE): # When absolute of dE is not equal to dE it means the scan is negative
                self.E = np.append(self.E, np.linspace(self.Eini, self.Elow, ntlower))
                self.E = np.append(self.E, np.linspace(self.Elow, self.Eupp, nt))
                self.E = np.append(self.E, np.linspace(self.Eupp, self.Eini, ntupper))
            elif (segments == 2) and (Eini == Eupp):
                self.E = np.append(self.E, np.linspace(self.Eini, self.Elow, nt))
                self.E = np.append(self.E, np.linspace(self.Elow, self.Eupp, nt))
            elif (segments == 2) and  (Eini == Elow):
                self.E = np.append(self.E, np.linspace(self.Eini, self.Eupp, nt))
                self.E = np.append(self.E, np.linspace(self.Eupp, self.Elow, nt))



class Step:
    """ 
    ----------
    Description:
    Returns t and E for a step potential waveform.
    All the parameters are given a default value.

    ----------
    Paramters:
    Es:     V, potential step
    ttot:   s, total time of the step
    dt:     s, time increment

    ----------
    Returns:
    self.t: s, time numpy array
    self.E: V, potential numpy array

    ----------
    Examples:
    > import softpotato as sp
    > ca = sp.technique.Step(Es=0.1) # Changing only the step potential
    > print(ca.t)
    > print(ca.E)
    """

    def __init__(self, Es=0.5, ttot=1, dt=0.01):
        
        self.Es = Es
        self.ttot = ttot
        self.dt = dt
        self.nt = int(self.ttot/self.dt)

        self.E = np.ones([self.nt])*self.Es
        self.t = np.linspace(dt, self.ttot + dt, self.nt)



class Custom:
    """ 
    
    Returns t and E for a customised potential waveform.
    
    Parameters
    ----------
    wf:     list containing the waveform object
    
    Returns
    -------
    t:      s, time array
    E:      V, potential array
    
    Examples
    --------
    >>> import softpotato as sp
    >>> wf1 = sp.step(Estep, tini, ttot, dt)
    >>> wf2 = sp.sweep(Eini, Efin, sr, dE, ns)
    >>> wf = sp.Construct_wf([wf1, wf2])
    
    Returns t and E calculated with the parameters given
    """

    def __init__(self, wf):
        n = len(wf)
        t = np.array([0])
        E = np.array([0])

        for i in range(n):
            t = np.concatenate([t,wf[i].t+t[-1]])
            E = np.concatenate([E,wf[i].E])

        # Remove first data point to prevent repeating time
        self.t = t[1:]
        self.E = E[1:]
        

if __name__ == '__main__':
    cv = Sweep2()
    print(cv.E)
    print(cv.t)



import numpy as np

F = 96485 # C/mol
R = 8.3145 # J/(mol K)


class Macro:
    '''
    '''

    def __init__(self, n=1, A=1, C=1e-6, D=1e-5, T=298, noise=0):
        self.n = n
        self.A = A
        self.C = C
        self.D = D
        self.T = 298
        self.noise = noise

    def Cottrell(self, t):
        i =  self.n*F*self.A*self.C*np.sqrt(self.D/(np.pi*t))
        return i + np.random.normal(size=t.size, scale=self.noise)

    def RandlesSevcik(self, sr):
        i = 0.4463*self.n*F*self.A*self.C*np.sqrt(self.n*F*sr*self.D/(R*self.T))
        return i + np.random.normal(size=sr.size, scale=self.noise)



class MicroDisc:
    '''
    ----------
    Description:
    Calculates the current of a microdisc electrode
    All parameters are given a default value

    ----------
    Parameters:
    a:      cm, electrode radius [5e-4 cm]
    n:      number of electrons [1]
    DO:     cm2/s, diffusion coefficient of species O [1e-5 cm2/s]
    DR:     cm2/s, diffusioen coefficient of species R [1e-5 cm2/s]
    CO:     mol/cm3, bulk concentration of species O [1e-6 mol/cm3]
    E0:     V, standard potential [0 V]
    k0:     cm/s, standard rate constant [1e8 cm/s]
    alpha:  transfer coefficient [0.5]
    T:      K, temperature [298 K]
    noise:  A, random normal distributed noise


    ----------
    Returns:
    self.iLim:  A, limiting current

    ----------
    Methods:
    LSV(E):
        Parameters:
        E:      V, required, numpy array to calculate the voltammogram
        Returns:
        self.lsv: A, numpy array with the currents calculated at potentials E
    CA():
        Parameters:
        t:      s, numpy array with the times to calculate the current
        E:      V, potential for the step
        Returns:
        self.ca: A, numpy array with the currents calculated at times t


    ----------
    Examples:
    > import softpotato as sp
    > E = sp.technique.Sweep()
    > disc = sp.calc.MicroDisc(a=12.5e-4) # Changing only the electrode radius
    > lsv = disc.voltammetry(E)
    > print(disc.iLim)
    > print(lsv)

    '''
    
    def __init__(self, a=5e-4, n=1, DO=1e-5, DR=1e-5, cOb=1e-6, cRb=0, E0=0, 
                 k0=1e8, alpha=0.5, T=298, noise=0):
        self.a = a
        self.n = n
        self.DO = DO
        self.DR = DR
        self.cOb = cOb
        self.cRb = cRb
        self.E0 = E0
        self.k0 = k0
        self.alpha = alpha
        self.T = T
        self.FRT = F/(R*T)
        self.noise = noise
        if self.cOb > 0 and self.cRb == 0:
            #print('Reduction')
            self.C = self.cOb
            self.D = self.DO
            self.alphaf = -alpha
            self.alphab = 1-alpha
            self.DODR = DO/DR
            self.sign = -1
        elif self.cOb == 0 and self.cRb > 0:
            #print('Oxidation')
            self.C = self.cRb
            self.D = self.DR
            self.DODR = DR/DO
            self.alphaf = -1+alpha
            self.alphab = alpha
            self.sign = 1
        else:
            print('Error. Either CO or CR should be zero.') 
        self.A = np.pi*a**2
        self.m0 = self.sign*4*self.D/(np.pi*self.a)
        self.km = m0
        self.iLim = n*F*self.A*self.m0*self.C

    def LSV(self, E):
        kf = self.k0*np.exp(self.alphaf*self.n*self.FRT*(E-self.E0))
        kb = self.k0*np.exp(self.alphab*self.n*self.FRT*(E-self.E0))
        if self.sign == -1:
            self.k = kf
        else:
            self.k = kb
        theta=1+self.DODR*np.exp(-self.sign*self.n*F*(E-self.E0)/(R*self.T))
        kappa = np.pi*self.k*self.a/(4*self.D)
        i = (self.iLim/theta)/(1 + 
            (np.pi/(kappa*theta))*((2*kappa*theta+
            3*np.pi)/(4*kappa*theta+3*np.pi**2)))
        return i + np.random.normal(size=E.size, scale=self.noise)

    def CA(self, t, E=-1):
        t = np.array(t)
        if self.sign == -1:
            fOfR = self.fun(t,self.DO)/self.fun(t,self.DR)
        else:
            fOfR = self.fun(t,self.DR)/self.fun(t,self.DO)
        _ = self.LSV(E)
        theta=1+self.DODR*fOfR*np.exp(-self.sign*self.n*F*(E-self.E0)/(R*self.T))
        kappa = self.k*self.a/(self.D*self.fun(t,self.D))
        self.im0 = self.sign*np.pi*self.n*F*self.D*self.C*self.a*self.fun(t,self.D)
        i = (self.im0/theta)/(1 + 
            (np.pi/(kappa*theta))*((2*kappa*theta+
            3*np.pi)/(4*kappa*theta+3*np.pi**2)))
        return i + np.random.normal(size=t.size, scale=self.noise)

    def fun(self, t, D):
        s = D*t/self.a**2
        f1 = 1/np.sqrt(np.pi*s) + 1 + np.sqrt(s/(4*np.pi)) - 3*s/25 + 3*s**(3/2)/226
        f2 = 4/np.pi + 8/np.sqrt(s*np.pi**5) + 25/(2792*s**(3/2)) - 1/(3880*s**(5/2)) - 1/(4500*s**(7/2))
        # Fancy if statement using boolean operations:
        return (s<1.281)*f1 + (s>=1.281)*f2


class MicroBand:
    '''
    '''

    def __init__(self, t=1, n=1, D=1e-5, cb=1e-6, w=5e-4, l=500e-4):
        self.A = w*l
        self.m0 = 2*np.pi*D/(w*log(64*D*t/w**2))
        self.km = m0
        self.iLim = n*F*self.A*self.m0*cb


if __name__ == '__main__':
    import plotting
    print('Running from main')
    noise = 0
    ox = 0
    if ox:
        disc = MicroDisc(cOb=0, cRb=1e-6, k0=1e-3, alpha=0.5, noise=noise)
    else:
        disc = MicroDisc(cOb=1e-6, cRb=0, k0=1e-3, alpha=0.5, noise=noise)
    print(disc.iLim)
    E = np.linspace(1,-1)
    t = np.linspace(0.001,1,1000)
    nt = t.size
    nE = E.size
    i = np.zeros([nt,nE])
    for x in range(nE):
        i[:,x] = disc.CA(t,E[x])
    plotting.plot(E,-i.T/i[:,-1],ylab='$i_{norm}$', show=1)




from setuptools import setup, find_packages, Extension

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='softpotato',
    version="1.0.14",
    license='GPL3',
    author='Oliver Rodriguez',
    author_email='oliver.rdz@softpotato.xyz',
    description='Open source electrochemistry toolkit',
    #long_description=long_description,
    #long_description_content_type='text/markdown',
    packages=find_packages('src'),
    package_dir={'':'src'},
    url='https://softpotato.xyz',
    keywords='Electrochemistry',
    python_requires=">=3.6",
    install_requires=[
        'numpy',
        'matplotlib',
        'scipy',
    ],
)
